import propar
import pandas as pd
import pathlib, os, json

homedir = pathlib.Path.home()
configfile = f'{homedir}/bronkhorstServerConfig/comConfg.log'

def getParamDF():
    paramDF = pd.DataFrame(columns=['dde_nr','proc_nr','parm_nr','parm_type'])
    db = propar.database().get_all_parameters()
    for dct in db:
        parmName = dct['parm_name']
        ddeNr = dct['dde_nr']
        procNr = dct['proc_nr']
        parmNr = dct['parm_nr']
        parmType = dct['parm_type']
        paramDF.loc[parmName] = [ddeNr,procNr,parmNr,parmType]
    return paramDF

def strToFloat(string):
    try:
        x = float(string)
        return x
    except:
        return string

paramDF = getParamDF()

def startMfc(com = 'COM1'):
    mfcMain = propar.instrument(com)
    return mfcMain


class MFC():
    def __init__(self,address, mfcMain):
        self.address = address
        self.mfcMain = mfcMain
        self.com = mfcMain.comport
        self.pollparams = ['User tag', 'Control mode', 'Fluid name', 'Fluidset index','fMeasure', 'fSetpoint', 
                  'Measure', 'Setpoint', 'Valve output']
        self.ddenrs = paramDF['dde_nr'].loc[self.pollparams].values
        self.getAddresses()
        #self.pollparamList = propar.database().get_parameters(self.ddenrs)
    def __str__(self):
        return self.readName()
    def getCom(self, com=None):
        if not com and os.path.exists(configfile):
            f = open(configfile,'r')
            comno = f.read()
            f.close()
            return f'COM{comno}'
        return com
    def getNumbers(self,name):
        proc_nr = paramDF.loc[name]['proc_nr']
        parm_nr = paramDF.loc[name]['parm_nr']
        parm_type = paramDF.loc[name]['parm_type']
        return proc_nr, parm_nr, parm_type
    def getddes(self,*names):
        return paramDF['dde_nr'].loc[list(names)].values

    def readParam(self,name, address = None):
        if address == None:
            address = self.address
        proc_nr, parm_nr, parm_type = self.getNumbers(name)
        parValue = self.mfcMain.master.read(address,proc_nr,parm_nr,parm_type)
        return parValue
    def readParams(self,ddeList, address=None):
        if not address:
            address = self.address
        paramdctlist = propar.database().get_parameters(ddeList)
        datalist = propar.instrument(self.com, address=address).read_parameters(paramdctlist)
        datadct = {}
        datadct['address'] = address
        for d in datalist:
            datadct[d['parm_name']] = d['data']
        return datadct
    def readParams_names(self, *names,address=None):
        ddes = self.getddes(*names)
        return self.readParams(ddes,address=address)
    def readParams_namesAllAddress(self,*names):
        datadct_all = {}
            
        for c,a in enumerate(self.addresses):
            dct = self.readParams_names(*names, address=a)
            for par in dct:
                if c == 0:
                    datadct_all[par] = []
                datadct_all[par].append(dct[par])
        return datadct_all
    
    def readParams_allAddsPars(self):
        return self.readParams_namesAllAddress(*self.pollparams)

    def writeParam(self,name, value):
        proc_nr, parm_nr, parm_type = self.getNumbers(name)
        x = self.mfcMain.master.write(self.address,proc_nr,parm_nr,parm_type,value)
        return x
    def writeSetpoint(self,value):
        name = self.readName()
        value = float(value)
        print(f'setting {name} to {value} ml/min')
        x =  self.writeParam('fSetpoint',value)
        return self.readSetpoint()
        
    def readSetpoint(self):
        sp = self.readParam('fSetpoint')
        name = self.readName()
        print(f'{name} setpoint {sp} ml/min')
        return sp
    def readFlow(self):
        flowRate = self.readParam('fMeasure')
        name = self.readName()
        print(f'{name} flow {flowRate} ml/min')
        return flowRate
    def readName(self):
        name = self.readParam('User tag')
        return name
    def writeName(self,name):
        x = self.writeParam('User tag',name)
        newname = self.readName()
        return newname
    def getAddresses(self):
        nodes = self.mfcMain.master.get_nodes()
        self.addresses = [n['address'] for n in nodes]
        addressesString = ' '.join([str(a) for a in self.addresses])
        return addressesString
    def readAddresses(self):
        return self.getAddresses()
    def readControlMode(self):
        mode = self.readParam('Control mode')
        name = self.readName()
        print(f'{name} control mode: {mode}')
        return mode
    def writeControlMode(self, value):
        value = int(value)
        x = self.writeParam('Control mode', value)
        newvalue = self.readControlMode()
        return newvalue
    def readFluidType(self):
        name = self.readName()
        fluiddct = self.readParams_names('Fluidset index', 'Fluid name')
        print(fluiddct)
        return fluiddct
    
    def writeFluidIndex(self,value):
        value = int(value)
        x = self.writeParam('Fluidset index',value)
        newIndex = self.readFluidType()
        return newIndex
    def readValve(self):
        value = self.readParam('Valve output')
        valve = value/2**24
        return valve
    def testMessage(self):
        return 'a'*1000 + 'b'*1000
    
    def pollAll(self):
        datadct = {}
        for par in ['address']+self.pollparams:
            datadct[par] = []
        for a in self.addresses:
            datadcttmp = self.readParams(self.ddenrs, a)
            for par in datadcttmp:
                datadct[par].append(datadcttmp[par])
        df = pd.DataFrame.from_dict(datadct)
        df['Measure'] = df['Measure'].apply(lambda x: x*100/32000)
        df['Setpoint'] = df['Setpoint'].apply(lambda x: x*100/32000)
        df['Valve output'] = df['Valve output'].apply(lambda x: x/2**24)
        df = df.rename({'Measure':'Measure_pct', 'Setpoint':'Setpoint_pct'}, axis = 1)
        return df
    def pollAllServer(self):
        df = self.pollAll()
        print(df)
        dfstring = ';'.join(df.columns)
        for i in df.index.values:
            dfstring += '\n'
            dfstring += ';'.join([str(x) for x in df.loc[i]])
        return dfstring
    
    def readMeasure_pct(self):
        m = self.readParam('Measure')
        m_pct = m*100/32000
        return m_pct
    def readSetpoint_pct(self):
        sp = self.readParam('Setpoint')
        sp_pct = sp*100/32000
        return sp_pct
    def writeSetpoint_pct(self,value_pct):
        value = int(value_pct*32000/100)
        self.writeParam('Setpoint', value)
        return self.readSetpoint_pct()
        
    def wink(self):
        if not self.com:
            print('com needs to be defined to run this')
            return
        return propar.instrument(self.com, self.address).wink()
    def strToMethod(self,inputString):
        stringSplit = inputString.split(';')
        #address = stringSplit[0]
        methodName = stringSplit[1]
        args = stringSplit[2:]
        methodDct = {'readName': self.readName, 'readParam':self.readParam,
                     'readSetpoint':self.readSetpoint, 'writeSetpoint':self.writeSetpoint,
                     'writeParam':self.writeParam, 'readFlow':self.readFlow,
                     'getAddresses': self.getAddresses, 'pollAll':self.pollAllServer,
                     'readControlMode': self.readControlMode, 'writeControlMode': self.writeControlMode,
                     'readFluidType':self.readFluidType, 'writeFluidIndex':self.writeFluidIndex,
                     'writeName':self.writeName, 'readMeasure_pct': self.readMeasure_pct,
                     'readSetpoint_pct': self.readSetpoint_pct, 'wink':self.wink,
                     'readValve': self.readValve, 'readParams_names':self.readParams_names,
                     'readParams_allAddsPars':self.readParams_allAddsPars,'testMessage':self.testMessage}
        method = methodDct[methodName]
        val = method(*args)
        if type(val) == dict:
            return json.dumps(val)
        return val
    

    



from PyQt6 import QtWidgets, QtCore, QtGui
import argparse
import time
import pandas as pd
import matplotlib.pyplot as plt

from .bronkhorstClient import MFCclient
from .bronkhorstServer import HOST, PORT, logdir
from .plotters import Plotter
from functools import partial
import logging
import pathlib, os, time
import socket

homedir = pathlib.Path.home()
fulllogdir = f'{homedir}/{logdir}'
os.makedirs(fulllogdir,exist_ok=True)
logger = logging.getLogger()


def parseArguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('-m','--maxMFCs', default=10, type=int, help='maximum number of MFCs that might be needed (default 10)')
    args = parser.parse_args()
    maxMFCs = args.maxMFCs
    return maxMFCs

class Worker(QtCore.QObject):
    outputs = QtCore.pyqtSignal(pd.DataFrame)
    def __init__(self, host, port, waittime = 1):
        super(Worker,self).__init__()
        self.host = host
        self.port = port
        self.waittime = waittime
        self.mfc = MFCclient(1,self.host,self.port, connid=f'{socket.gethostname()}GUIthread')
        self.running = True
    def run(self):
        while self.running:
            try:
                df = self.mfc.pollAll()
            except (OSError, AttributeError, ConnectionResetError):
                message = "connection to server lost. Stopping polling"
                print(message)
                logger.warning(message)
                self.outputs.emit(pd.DataFrame())
                return
            except Exception as e:
                logger.exception(e)
                raise e
            self.outputs.emit(df)
            #QtCore.QThread.msleep(int(self.waittime*1000))
            time.sleep(self.waittime)
        print('stopping polling')

    def stop(self):
        self.running = False

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        eventlogfile = f'{homedir}/{logdir}/mfcgui.log'
        logging.basicConfig(filename=eventlogfile, level = logging.INFO, format = '%(asctime)s %(levelname)-8s %(message)s',
                            datefmt = '%Y/%m/%d_%H:%M:%S')
        logger.info('mfcgui opened')
        self.connid = f'{socket.gethostname()}GUI'
        self.MainWindow = MainWindow
        self.MainWindow.setObjectName("MainWindow")
        self.MainWindow.resize(800, 550)
        self.MainWindow.setWindowTitle('Bronkhorst GUI')
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.maxMFCs = parseArguments()
        self.box1x = 70
        self.box1y = 20
        self.xspacing = 90
        self.yspacing = 25

        spinboxsizex = 100

        rows = {'wink':0,
                'address':1,
                'setpoint':2,
                'measure':3,
                'setpointpct':4,
                'measurepct':5,
                'valve':6,
                'controlMode':7,
                'fluidIndex':8,
                'fluidName':9,
                'writesp':10,
                'usertag':11}

        self.scrollArea = QtWidgets.QScrollArea()
        self.scrollArea.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        
        self.outerLayout = QtWidgets.QVBoxLayout()

        self.group = QtWidgets.QGroupBox()

        self.topLayout = QtWidgets.QHBoxLayout()
        self.gridLayout = QtWidgets.QGridLayout()
        self.leftLayout = QtWidgets.QGridLayout()
        

        self.bottomLayout = QtWidgets.QGridLayout()
        self.bottomLayout.setVerticalSpacing(0)
        
        
        self.startButton = QtWidgets.QPushButton()
        self.startButton.setObjectName('startButton')
        self.startButton.setMinimumWidth(150)
        self.startButton.setText('connect MFCs')
        self.bottomLayout.addWidget(self.startButton,0,0)

        self.runningIndicator = QtWidgets.QRadioButton()
        self.runningIndicator.setObjectName('runningIndicator')
        self.runningIndicator.setText('blinks when running')
        self.runningIndicator.setChecked(False)
        self.bottomLayout.addWidget(self.runningIndicator,1,0)

        self.plotBox = QtWidgets.QCheckBox()
        self.plotBox.setObjectName('plotBox')
        self.plotBox.setText('plot data?')
        self.plotBox.setEnabled(True)
        self.bottomLayout.addWidget(self.plotBox)

        self.hostInput = QtWidgets.QLineEdit()
        self.hostInput.setObjectName('hostInput')
        self.hostInput.setMinimumWidth(120)
        self.hostInput.setText(HOST)
        self.bottomLayout.addWidget(self.hostInput, 0, 1)

        self.hostLabel = QtWidgets.QLabel()
        self.hostLabel.setObjectName('hostLabel')
        self.hostLabel.setText('host name')
        self.bottomLayout.addWidget(self.hostLabel,1,1)


        self.portInput = QtWidgets.QSpinBox()
        self.portInput.setObjectName('portInput')
        self.portInput.setMinimumWidth(120)
        self.portInput.setMaximum(2**16)
        self.portInput.setMinimum(8000)
        self.portInput.setValue(PORT)
        self.bottomLayout.addWidget(self.portInput,0,2)

        self.portLabel = QtWidgets.QLabel()
        self.portLabel.setObjectName('portLabel')
        self.portLabel.setText('port value')
        self.bottomLayout.addWidget(self.portLabel,1,2)

        self.pollTimeBox = QtWidgets.QDoubleSpinBox()
        self.pollTimeBox.setObjectName('pollTimeBox')
        self.pollTimeBox.setValue(0.5)
        self.pollTimeBox.setMinimum(0.1)
        self.pollTimeBox.setMaximum(5)
        self.pollTimeBox.setDecimals(1)
        self.pollTimeBox.setSingleStep(0.1)
        self.bottomLayout.addWidget(self.pollTimeBox,0,3)

        self.pollLabel = QtWidgets.QLabel()
        self.pollLabel.setObjectName('pollLabel')
        self.pollLabel.setText('poll time')
        self.bottomLayout.addWidget(self.pollLabel,1,3)

        self.winkLabel = QtWidgets.QLabel()
        self.winkLabel.setObjectName('winkLabel')
        self.winkLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.winkLabel,rows['wink'],0)

        self.addressLabel = QtWidgets.QLabel()
        self.addressLabel.setObjectName('addressLabel')
        self.addressLabel.setText('address')
        self.addressLabel.adjustSize()
        self.addressLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.addressLabel, rows['address'],0)

        self.spLabel = QtWidgets.QLabel()
        self.spLabel.setObjectName('spLabel')
        self.spLabel.setText('setpoint')
        self.spLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.spLabel,rows['setpoint'],0)

        self.measureLabel = QtWidgets.QLabel()
        self.measureLabel.setObjectName('measureLabel')
        self.measureLabel.setText('measure')
        self.measureLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.measureLabel,rows['measure'],0)

        self.sppctLabel = QtWidgets.QLabel()
        self.sppctLabel.setObjectName('sppctLabel')
        self.sppctLabel.setText('setpoint(%)')
        self.sppctLabel.adjustSize()
        self.sppctLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.sppctLabel,rows['setpointpct'],0)

        self.measurepctLabel = QtWidgets.QLabel()
        self.measurepctLabel.setObjectName('measurepctLabel')
        self.measurepctLabel.setText('measure(%)')
        self.measurepctLabel.adjustSize()
        self.measurepctLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.measurepctLabel,rows['measurepct'],0)

        self.valveLabel = QtWidgets.QLabel()
        self.valveLabel.setObjectName('valveLabel')
        self.valveLabel.setText('valve output')
        self.valveLabel.adjustSize()
        self.valveLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.valveLabel,rows['valve'],0)

        self.controlLable = QtWidgets.QLabel()
        self.controlLable.setObjectName('controlLable')
        self.controlLable.setText('control mode')
        self.controlLable.adjustSize()
        self.controlLable.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.controlLable,rows['controlMode'],0)

        self.fluidLabel = QtWidgets.QLabel()
        self.fluidLabel.setObjectName('fluidLabel')
        self.fluidLabel.setText('fluid index')
        self.fluidLabel.adjustSize()
        self.fluidLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.fluidLabel,rows['fluidIndex'],0)

        self.fluidNameLabel = QtWidgets.QLabel()
        self.fluidNameLabel.setObjectName('fluidNameLabel')
        self.fluidNameLabel.setText('fluid name')
        self.fluidNameLabel.adjustSize()
        self.fluidNameLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.fluidNameLabel,rows['fluidName'],0)

        self.writespLabel = QtWidgets.QLabel()
        self.writespLabel.setObjectName('writespLabel')
        self.writespLabel.setText('write setpoint')
        self.writespLabel.adjustSize()
        self.writespLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.writespLabel,rows['writesp'],0)

        self.userTagLabel = QtWidgets.QLabel()
        self.userTagLabel.setObjectName('userTagLabel')
        self.userTagLabel.setText('user tag')
        self.userTagLabel.adjustSize()
        self.userTagLabel.setMinimumHeight(self.yspacing)
        self.leftLayout.addWidget(self.userTagLabel,rows['usertag'],0)

        self.controlModeDct = {
                "0;Bus/RS232":0 ,
                "1;Analog input":1 ,
                "2;FB/RS232 slave":2 ,
                "3;Valve close":3 ,
                "4;Controller idle":4 ,
                "5;Testing mode":5 ,
                "6;Tuning mode":6 ,
                "7;Setpoint 100%":7 ,
                "8;Valve fully open":8 ,
                "9;Calibration mode":9 ,
                "10;Analog slave":10,
                "11;Keyb. & FLOW-BUS":11,
                "12;Setpoint 0%":12,
                "13;FB, analog slave":13,
                "14;(FPP) Range select":14,
                "15;(FPP) Man.s, auto.e":15,
                "16;(FPP) Auto.s, man.e":16,
                "17;(FPP) Auto.s, auto.e":17,
                "18;RS232":18,
                "19;RS232 broadcast":19,
                "20;Valve steering":20,
                "21;An. valve steering":21}
        
        self.winkbuttons= {}
        self.addressLabels = {}
        self.setpointBoxes = {}
        self.measureBoxes = {}
        self.setpointpctBoxes = {}
        self.measurepctBoxes = {}
        self.valveBoxes = {}
        self.controlBoxes = {}
        self.fluidBoxes = {}
        self.fluidNameBoxes = {}
        self.writeSetpointBoxes = {}
        self.writeSetpointpctBoxes = {}
        self.userTags = {}

        self.running = False
        for i in range(self.maxMFCs):
            self.winkbuttons[i] = QtWidgets.QPushButton()
            self.winkbuttons[i].setText('wink')
            self.winkbuttons[i].setObjectName(f'winkbuttons{i}')
            self.winkbuttons[i].setMaximumWidth(spinboxsizex)
            self.winkbuttons[i].setMinimumHeight(self.yspacing)
            self.winkbuttons[i].setEnabled(False)
            self.gridLayout.addWidget(self.winkbuttons[i], rows['wink'],i+1)
            self.winkbuttons[i].clicked.connect(partial(self.wink,i))

            self.addressLabels[i] = QtWidgets.QSpinBox()
            self.addressLabels[i].setObjectName(f'addressLabel{i}')
            self.addressLabels[i].setMinimum(-1)
            self.addressLabels[i].setMaximum(99)
            self.addressLabels[i].setValue(-1)
            self.addressLabels[i].setMaximumWidth(spinboxsizex)
            self.addressLabels[i].setMinimumHeight(self.yspacing)
            self.addressLabels[i].setEnabled(False)
            self.gridLayout.addWidget(self.addressLabels[i], rows['address'], i+1)

            self.setpointBoxes[i] = QtWidgets.QDoubleSpinBox()
            self.setpointBoxes[i].setObjectName(f'setpointBox{i}')
            self.setpointBoxes[i].setEnabled(False)
            self.setpointBoxes[i].setKeyboardTracking(False)
            self.setpointBoxes[i].setStyleSheet('color: black;')
            self.setpointBoxes[i].setMaximum(200)
            self.setpointBoxes[i].setMinimumHeight(self.yspacing)
            self.setpointBoxes[i].setMaximumWidth(spinboxsizex)
            self.gridLayout.addWidget(self.setpointBoxes[i], rows['setpoint'], i+1)

            self.measureBoxes[i] = QtWidgets.QDoubleSpinBox()
            self.measureBoxes[i].setObjectName(f'measureBox{i}')
            self.measureBoxes[i].setEnabled(False)
            self.measureBoxes[i].setStyleSheet('color: black;')
            self.measureBoxes[i].setMaximum(200)
            self.measureBoxes[i].setMinimumHeight(self.yspacing)
            self.measureBoxes[i].setMaximumWidth(120)
            self.gridLayout.addWidget(self.measureBoxes[i], rows['measure'],i+1)

            self.setpointpctBoxes[i] = QtWidgets.QDoubleSpinBox()
            self.setpointpctBoxes[i].setObjectName(f'setpointpctBox{i}')
            self.setpointpctBoxes[i].setEnabled(False)
            self.setpointpctBoxes[i].setStyleSheet('color: black;')
            self.setpointpctBoxes[i].setMaximum(200)
            self.setpointpctBoxes[i].setMinimumHeight(self.yspacing)
            self.setpointpctBoxes[i].setMaximumWidth(spinboxsizex)
            self.gridLayout.addWidget(self.setpointpctBoxes[i],rows['setpointpct'],i+1)

            self.measurepctBoxes[i] = QtWidgets.QDoubleSpinBox()
            self.measurepctBoxes[i].setObjectName(f'measurepctBox{i}')
            self.measurepctBoxes[i].setEnabled(False)
            self.measurepctBoxes[i].setStyleSheet('color: black;')
            self.measurepctBoxes[i].setMaximum(200)
            self.measurepctBoxes[i].setMinimumHeight(self.yspacing)
            self.measurepctBoxes[i].setMaximumWidth(spinboxsizex)
            self.gridLayout.addWidget(self.measurepctBoxes[i], rows['measurepct'],i+1)

            self.valveBoxes[i] = QtWidgets.QDoubleSpinBox()
            self.valveBoxes[i].setObjectName(f'valveBox{i}')
            self.valveBoxes[i].setEnabled(False)
            self.valveBoxes[i].setStyleSheet('color: black;')
            self.valveBoxes[i].setMaximumWidth(spinboxsizex)
            self.valveBoxes[i].setMinimumHeight(self.yspacing)
            self.gridLayout.addWidget(self.valveBoxes[i], rows['valve'],i+1)

            self.controlBoxes[i] = QtWidgets.QComboBox()
            self.controlBoxes[i].setObjectName(f'controlBoxes{i}')
            self.controlBoxes[i].setEnabled(False)
            #self.controlBoxes[i].setStyleSheet('color: black;')
            self.controlBoxes[i].setMaximumWidth(spinboxsizex)
            self.controlBoxes[i].setMinimumHeight(self.yspacing)
            #for mode in self.controlModeDct:
            #    self.controlBoxes[i].addItem(mode)
            self.controlBoxes[i].addItem("0;Bus/RS232")
            self.controlBoxes[i].addItem("1;Analog input")
            self.controlBoxes[i].addItem("2;FB/RS232 slave")
            self.controlBoxes[i].addItem("3;Valve close")
            self.controlBoxes[i].addItem("4;Controller idle")
            self.controlBoxes[i].addItem("5;Testing mode")
            self.controlBoxes[i].addItem("6;Tuning mode")
            self.controlBoxes[i].addItem("7;Setpoint 100%")
            self.controlBoxes[i].addItem("8;Valve fully open")
            self.controlBoxes[i].addItem("9;Calibration mode")
            self.controlBoxes[i].addItem("10;Analog slave")
            self.controlBoxes[i].addItem("11;Keyb. & FLOW-BUS")
            self.controlBoxes[i].addItem("12;Setpoint 0%")
            self.controlBoxes[i].addItem("13;FB, analog slave")
            self.controlBoxes[i].addItem("14;(FPP) Range select")
            self.controlBoxes[i].addItem("15;(FPP) Man.s, auto.e")
            self.controlBoxes[i].addItem("16;(FPP) Auto.s, man.e")
            self.controlBoxes[i].addItem("17;(FPP) Auto.s, auto.e")
            self.controlBoxes[i].addItem("18;RS232")
            self.controlBoxes[i].addItem("19;RS232 broadcast")
            self.controlBoxes[i].addItem("20;Valve steering")
            self.controlBoxes[i].addItem("21;An. valve steering")
            self.controlBoxes[i].currentIndexChanged.connect(partial(self.setControlMode,i))
            self.gridLayout.addWidget(self.controlBoxes[i], rows['controlMode'],i+1)

            self.fluidBoxes[i] = QtWidgets.QSpinBox()
            self.fluidBoxes[i].setObjectName(f'fluidBoxes{i}')
            self.fluidBoxes[i].setEnabled(False)
            self.fluidBoxes[i].setStyleSheet('color: black;')
            self.fluidBoxes[i].setMaximum(20)
            self.fluidBoxes[i].setMinimumHeight(self.yspacing)
            self.fluidBoxes[i].setMaximumWidth(spinboxsizex)
            self.fluidBoxes[i].setKeyboardTracking(False)
            self.fluidBoxes[i].valueChanged.connect(partial(self.setFluidIndex,i))
            self.gridLayout.addWidget(self.fluidBoxes[i], rows['fluidIndex'],i+1)

            self.fluidNameBoxes[i] = QtWidgets.QLabel()
            self.fluidNameBoxes[i].setObjectName(f'fluidNameBoxes{i}')
            self.fluidNameBoxes[i].setMinimumHeight(self.yspacing)
            self.fluidNameBoxes[i].setMaximumWidth(spinboxsizex)
            self.gridLayout.addWidget(self.fluidNameBoxes[i], rows['fluidName'],i+1)

            self.writeSetpointBoxes[i] = QtWidgets.QDoubleSpinBox()
            self.writeSetpointBoxes[i].setObjectName(f'writeSetpointBox{i}')
            self.writeSetpointBoxes[i].setEnabled(False)
            self.writeSetpointBoxes[i].setStyleSheet('color: black;')
            self.writeSetpointBoxes[i].setMaximum(200)
            self.writeSetpointBoxes[i].setKeyboardTracking(False)
            self.writeSetpointBoxes[i].valueChanged.connect(partial(self.setFlow, i))
            self.writeSetpointBoxes[i].setMaximumWidth(spinboxsizex)
            self.writeSetpointBoxes[i].setMinimumHeight(self.yspacing)
            self.gridLayout.addWidget(self.writeSetpointBoxes[i],rows['writesp'],i+1)

            self.userTags[i] = QtWidgets.QLineEdit()
            self.userTags[i].setObjectName(f'userTag{i}')
            self.userTags[i].setEnabled(False)
            self.userTags[i].returnPressed.connect(partial(self.setUserTag, i))
            self.userTags[i].setMaximumWidth(spinboxsizex)
            self.userTags[i].setMinimumHeight(self.yspacing)
            self.gridLayout.addWidget(self.userTags[i],rows['usertag'],i+1)
 
        self.group.setLayout(self.gridLayout)
        
        self.scrollArea.setWidget(self.group)
        self.scrollArea.setMinimumHeight(self.yspacing*(len(rows)+4))

        self.leftLayout.setVerticalSpacing(0)
        self.scrollArea3 = QtWidgets.QScrollArea()
        self.scrollArea3.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.scrollArea3.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        

        self.group3 = QtWidgets.QGroupBox()
        self.group3.setLayout(self.leftLayout)
        
        self.group3.setFixedHeight(self.group.height())

        self.scrollArea3.setWidget(self.group3)
        self.scrollArea3.setFixedWidth(self.group3.width())

        self.topLayout.addWidget(self.scrollArea3)
        self.topLayout.addWidget(self.scrollArea)
        
        self.topLayout.setSpacing(0)
        self.outerLayout.addLayout(self.topLayout)
        
        self.group2 = QtWidgets.QGroupBox()
        self.group2.setLayout(self.bottomLayout)
        self.scrollArea2 = QtWidgets.QScrollArea()
        self.scrollArea2.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.scrollArea2.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.scrollArea2.setWidget(self.group2)

        self.outerLayout.addWidget(self.scrollArea2)
        self.centralwidget.setLayout(self.outerLayout)
        
        self.MainWindow.setCentralWidget(self.centralwidget)
        QtCore.QMetaObject.connectSlotsByName(self.MainWindow)


        self.startButton.clicked.connect(self.connectLoop)

    def connectMFCs(self):
        self.host = self.hostInput.text()
        self.port = self.portInput.value()
        try:
            df = MFCclient(1,self.host,self.port, connid=self.connid).pollAll()
        except OSError as e:
            raise OSError(e)
        self.plot = self.plotBox.isChecked()
        self.plotBox.setEnabled(False)
        if self.plot:
            plt.ion()
            self.plotter = Plotter(host = self.host, port = self.port)
        
        self.enabledMFCs = []
        self.originalUserTags = {}
        self.originalControlModes = {}
        self.originalFluidIndexes = {}
        self.originalSetpoints = {}
        for i in df.index.values:
            self.originalSetpoints[i] = df.loc[i]['fSetpoint']
            self.writeSetpointBoxes[i].setValue(self.originalSetpoints[i])
            self.enabledMFCs.append(i)
            self.originalUserTags[i] = df.loc[i]['User tag']
            self.originalControlModes[i] = df.loc[i]['Control mode']
            self.userTags[i].setText(self.originalUserTags[i])
            self.controlBoxes[i].setCurrentIndex(self.originalControlModes[i])
            self.originalFluidIndexes[i] = df.loc[i]['Fluidset index']
            self.fluidBoxes[i].setValue(self.originalFluidIndexes[i])
        self.updateMFCs(df)
        logger.info(f'connected to server. Host: {self.host}, port: {self.port}')

    def updateMFCs(self,df):
        if len(df.columns) == 0:
            self.disableWidgets()
            return
        checkValue = self.runningIndicator.isChecked()
        self.runningIndicator.setChecked(not checkValue)
        for i in df.index.values:
            try:
                self.winkbuttons[i].setEnabled(True)
                newSetpoint = df.loc[i]['fSetpoint']
                newControlMode = df.loc[i]['Control mode']
                newFluidIndex = df.loc[i]['Fluidset index']
                self.addressLabels[i].setValue(df.loc[i]['address'])
                self.addressLabels[i].setStyleSheet('color: black;')
                self.setpointBoxes[i].setValue(newSetpoint)
                self.measureBoxes[i].setValue(df.loc[i]['fMeasure'])
                self.valveBoxes[i].setValue(df.loc[i]['Valve output'])
                self.setpointpctBoxes[i].setValue(df.loc[i]['Setpoint_pct'])
                self.measurepctBoxes[i].setValue(df.loc[i]['Measure_pct'])
                self.writeSetpointBoxes[i].setEnabled(True)
                self.controlBoxes[i].setEnabled(True)
                self.fluidBoxes[i].setEnabled(True)
                self.fluidNameBoxes[i].setText(df.loc[i]['Fluid name'])
                
                if newControlMode != self.originalControlModes[i]:
                    self.controlBoxes[i].setCurrentIndex(newControlMode)
                    self.originalControlModes[i] = newControlMode
                
                if newFluidIndex != self.originalFluidIndexes[i]:
                    self.fluidBoxes[i].setValue(newFluidIndex)
                    self.originalFluidIndexes[i] = newFluidIndex
                newUserTag = df.loc[i]['User tag']
                if newUserTag != self.originalUserTags[i]:
                    self.userTags[i].setText(df.loc[i]['User tag'])
                    self.originalUserTags[i] = newUserTag
                if newSetpoint != self.originalSetpoints[i]:
                    self.writeSetpointBoxes[i].setValue(newSetpoint)
                    self.originalSetpoints[i] = newSetpoint
                self.userTags[i].setEnabled(True)
            except TypeError as e:
                print(df)
                logger.warning(e)
                return
            except Exception as e:
                print(df)
                logger.exception(e)
                raise e
        
        
        if self.plot and self.running:
            self.plotter.plotAllSingle(df)
        
    def connectLoop(self):
        if not self.running:
            self.host = self.hostInput.text()
            self.port = self.portInput.value()
            self.waittime = self.pollTimeBox.value()
            try:
                self.connectMFCs()
            except OSError:
                message = f"couldn't find server at host: {self.host}, port: {self.port}. Try starting it or checking host and port settings"
                print(message)
                logger.warning(message)
                return
            except Exception as e:
                logger.exception(e)
                raise e
            self.running = True
            self.startButton.setText('stop connection')
            self.worker = Worker(self.host,self.port, self.waittime)
            self.thread = QtCore.QThread()
            self.worker.moveToThread(self.thread)
            self.thread.started.connect(self.worker.run)
            self.worker.outputs.connect(self.updateMFCs)
            self.thread.start()
        else:
            self.stopConnect()
            self.disableWidgets()
            logger.info(f'connection closed to server at host: {self.host}, port {self.port}')

    def stopConnect(self):
        self.worker.stop()
        self.thread.quit()
        self.running = False
        self.worker.deleteLater()
        self.plotBox.setEnabled(True)
        if self.plot:
            plt.close()
    
    def disableWidgets(self):
        self.running = False
        self.startButton.setText('connect MFCs')
        for i in range(self.maxMFCs):
            self.writeSetpointBoxes[i].setEnabled(False)
            self.controlBoxes[i].setEnabled(False)
            self.fluidBoxes[i].setEnabled(False)
            self.userTags[i].setEnabled(False)
            self.winkbuttons[i].setEnabled(False)

    def setFlow(self,i):
        if not self.running:
            return
        value = self.writeSetpointBoxes[i].value()
        address = self.addressLabels[i].value()
        print(f'setting flow to {value} on address {address}')
        newflow = MFCclient(address,self.host, self.port, connid=self.connid).writeSetpoint(value)
        self.writeSetpointBoxes[i].setValue(newflow)
        
    def setUserTag(self,i):
        if not self.running:
            return
        value = self.userTags[i].text()
        address = self.addressLabels[i].value()
        print(f'setting flow to {value} on address {address}')
        newtag = MFCclient(address,self.host, self.port, connid=self.connid).writeName(value)
        self.userTags[i].setText(newtag)

    def setFlowAll(self):
        if not self.running:
            return
        for i in self.enabledMFCs:
            self.setFlow(i)

    def setControlMode(self,i):
        if not self.running:
            return
        value = self.controlBoxes[i].currentIndex()
        #text = self.controlBoxes[i].currentText()
        #value = int(text.split(';')[0])
        address = self.addressLabels[i].value()
        print(f'setting address {address} to control mode {value}')
        
        newmode = MFCclient(address, self.host,self.port, connid=self.connid).writeControlMode(value)
        self.controlBoxes[i].setCurrentIndex(newmode)


    def setFluidIndex(self,i):
        if not self.running:
            return
        value = self.fluidBoxes[i].value()
        address = self.addressLabels[i].value()
        print(f'setting address {address} to fluid {value}')
        newfluid = MFCclient(address,self.host,self.port, connid=self.connid).writeFluidIndex(value)
        newfluidIndex = newfluid['Fluidset index']
        self.fluidBoxes[i].setValue(newfluidIndex)
    
    def wink(self,i):
        address = self.addressLabels[i].value()
        MFCclient(address,self.host,self.port, connid=self.connid).wink()


def main():
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec())



import socket
import pandas as pd
import selectors,types
import os, pathlib
from bronkhorstControlbm31.bronkhorstServer import PORT, HOST, logdir
import json
import logging
import numpy as np

homedir = pathlib.Path.home()
#logdir = 'bronkhorstLogger'
fulllogdir = f'{homedir}/{logdir}'
os.makedirs(fulllogdir,exist_ok=True)
logger = logging.getLogger()




def connect(host=HOST, port=PORT):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host,port))
    return s

class MFCclient():
    def __init__(self,address, host=HOST,port=PORT, connid = socket.gethostname(),multi=True):
        self.address = address
        self.host = host
        self.port = port
        self.connid = connid
        self.multi = multi
        self.types = {'fMeasure': float, 'address':np.uint8, 'fSetpoint':float, 'Setpoint_pct':float, 'Measure_pct':float, 'Valve output': float, 
                 'Fluidset index': np.uint8,  'Control mode':np.uint8}
    def strToBool(self,string):
        if string == 'True' or string == 'False':
            return string == 'True'
        return string
    def readAddresses(self):
        string = self.makeMessage(self.address, 'getAddresses')
        addressesString = self.sendMessage(string)
        addresses = [int(a) for a in addressesString.split()]
        self.addresses = addresses
        print(addresses)
        return addresses
    def readName(self):
        string = self.makeMessage(self.address, 'readName')
        data = self.sendMessage(string)
        return data
    def writeName(self,newname):
        string = self.makeMessage(self.address,'writeName',newname)
        data = self.sendMessage(string)
        return data
    def readParam(self, name):
        string = self.makeMessage(self.address, 'readParam', name)
        data = self.sendMessage(string)
        return data
    def readParams(self,*names):
        string = self.makeMessage(self.address, 'readParams_names', *names)
        data = self.sendMessage(string)
        datadct = json.loads(data)
        return datadct
    def readFlow(self):
        string = self.makeMessage(self.address, 'readFlow')
        data = self.sendMessage(string)
        return float(data)
    def readSetpoint(self):
        string = self.makeMessage(self.address, 'readSetpoint')
        data = self.sendMessage(string)
        return float(data)
    def writeParam(self, name, value):
        string = self.makeMessage(self.address, 'writeParam', name, value)
        data = self.sendMessage(string)
        return self.strToBool(data)
    def writeSetpoint(self,value):
        string = self.makeMessage(self.address, 'writeSetpoint', value)
        data = self.sendMessage(string)
        return float(data)
    def readControlMode(self):
        string = self.makeMessage(self.address, 'readControlMode')
        data = self.sendMessage(string)
        return int(data)
    def writeControlMode(self,value):
        string = self.makeMessage(self.address, 'writeControlMode',value)
        data = self.sendMessage(string)
        return int(data)
    def readFluidType(self):
        string = self.makeMessage(self.address, 'readFluidType')
        data = self.sendMessage(string)
        datadct = json.loads(data)
        return datadct
    def writeFluidIndex(self,value):
        string = self.makeMessage(self.address, 'writeFluidIndex',value)
        data = self.sendMessage(string)
        return json.loads(data)
    def readMeasure_pct(self):
        string = self.makeMessage(self.address,'readMeasure_pct')
        data = self.sendMessage(string)
        return float(data)
    def readSetpoint_pct(self):
        string = self.makeMessage(self.address,'readSetpoint_pct')
        data = self.sendMessage(string)
        return float(data)
    def readValve(self):
        string = self.makeMessage(self.address,'readValve')
        data = self.sendMessage(string)
        return float(data)

    def pollAll(self):
        string = self.makeMessage(self.address, 'pollAll')
        data = self.sendMessage(string)
        datalines = data.split('\n')
        columns = datalines[0].split(';')
        array = [[i for i in line.split(';')] for line in datalines[1:] if line]
        df = pd.DataFrame(data = array,columns=columns)
        df = df.astype(self.types)
        return df
    
    def pollAll2(self):
        string = self.makeMessage(self.address,'readParams_allAddsPars')
        data = self.sendMessage(string)
        datadct = json.loads(data)
        df = pd.DataFrame.from_dict(datadct)
        df['Measure'] = df['Measure'].apply(lambda x: x*100/32000)
        df['Setpoint'] = df['Setpoint'].apply(lambda x: x*100/32000)
        df['Valve output'] = df['Valve output'].apply(lambda x: x/2**24)
        df = df.rename({'Measure':'Measure_pct', 'Setpoint':'Setpoint_pct'}, axis = 1)
        df = df.astype(self.types)
        return df

    def testMessage(self):
        string = self.makeMessage(self.address,'testMessage')
        data = self.sendMessage(string)
        return data

    def wink(self):
        string = self.makeMessage(self.address,'wink')
        data = self.sendMessage(string)
        return data
    def sendMessage(self,message):
        bytemessage = bytes(message,encoding='utf-8')
        if not self.multi:
            self.s = connect(self.host,self.port)
            self.s.sendall(bytemessage)
            data = self.s.recv(1024)
            self.s.close()
            strdata = data.decode()
            strdata = strdata.replace('!','')
        else:
            strdata = self.multiClient(bytemessage)
        print(strdata)
        return strdata
    def makeMessage(self, *args):
        sep = ';'
        string = f'{args[0]}'
        for arg in args[1:]:
            string += f'{sep}{arg}'
        return string

    def multiClient(self,message):
        sel = selectors.DefaultSelector()
        server_addr = (self.host, self.port)
        print(f"Starting connection {self.connid} to {server_addr}")
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(5)
        #sock.setblocking(False)
        sock.connect_ex(server_addr)
        events = selectors.EVENT_READ | selectors.EVENT_WRITE
        data = types.SimpleNamespace(
            connid=self.connid,
            msg_total=len(message),
            recv_total=0,
            messages=[message],
            outb=b"",
        )
        sel.register(sock, events, data=data)
        try:
            while True:
                events = sel.select(timeout=1)
                if events:
                    for key, mask in events:
                        receivedMessage = self.service_connection(key, mask,sel)
                        if receivedMessage:
                            receivedMessage = receivedMessage.replace('!','')
                # Check for a socket being monitored to continue.
                if not sel.get_map():
                    break
        except KeyboardInterrupt:
            print("Caught keyboard interrupt, exiting")
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            sel.close()
        return receivedMessage

    def service_connection(self,key, mask,sel):
        sock = key.fileobj
        data = key.data
        receivedMessage = b''
        strMessage = '!'
        if mask & selectors.EVENT_READ:
            while True:
                recv_data = sock.recv(1024)  # Should be ready to read
                if recv_data:
                    #print(f"Received {recv_data!r} from connection {data.connid}")
                    receivedMessage+= recv_data
                    data.recv_total += len(recv_data)
                    if receivedMessage:
                        strMessage = receivedMessage.decode()
                if not recv_data or '!' in strMessage:
                    print(f"Closing connection {data.connid}")
                    sel.unregister(sock)
                    sock.close()
                    if recv_data:
                        return strMessage
                    return
                
            
        if mask & selectors.EVENT_WRITE:
            if not data.outb and data.messages:
                data.outb = data.messages.pop(0)
            if data.outb:
                print(f"Sending {data.outb} to connection {data.connid}")
                sent = sock.send(data.outb)  # Should be ready to write
                data.outb = data.outb[sent:]




import socket
from bronkhorstControlbm31.bronkhorst import MFC, startMfc, configfile
import os, pathlib
import argparse
import selectors, types
import subprocess
import logging

logger = logging.getLogger()
homedir = pathlib.Path.home()
logdir = 'bronkhorstLogger'
fulllogdir = f'{homedir}/{logdir}'
os.makedirs(fulllogdir,exist_ok=True)
logfile = f'{homedir}/{logdir}/bronkhorstServer.log'


HOST = 'localhost'
PORT = 61245
com = 'COM1'

def getIP():
    p = subprocess.run('ipconfig',capture_output=True,text=True)
    ipLine = [l for l in p.stdout.split('\n') if 'IPv4 Address' in l][0]
    ip = ipLine.split()[-1]
    return ip

def isValidIP(ipaddress):
    try:
        hostname = socket.gethostbyaddr(ipaddress)[0].split('.')[0]
        return hostname.lower() == socket.gethostname().lower()
    except:
        return False
    

if not os.path.exists(os.path.dirname(configfile)):
    os.makedirs(os.path.dirname(configfile))

def getArgs(port=PORT):
    parser = argparse.ArgumentParser()
    parser.add_argument('host',nargs='?', default='local', help = ('can be "local" (localhost), "remote" (host name), '
                                            '"remoteip" (ip address), or the IP address of the host. The last two options are '
                                            'due to complications which may arise from a host having multiple connections'))

    if not os.path.exists(configfile):
        defaultCom = 1
    else:
        f = open(configfile,'r')
        defaultCom = f.read()
        f.close()
    parser.add_argument('-c','--com', default=defaultCom, help=('COM port (as number, e.g. -c 1) where MFCs are connected to. '
                                                                'If unsure, check in Device Manager under Ports. Default is last used'))
    parser.add_argument('-p','--port',default=port, type=int, help= ('port number. There is a default, but you can choose. Anything from '
                                                                '49152-65535 should be fine'))
    parser.add_argument('-a','--accepted-hosts', default = None, type = str, help= ('list of comma separated hosts that can be'
                                                                                    'accepted (default - accept all). E.g. -a pc1,pc2'))
    parser.add_argument('-d','--debug', action='store_true')

    args = parser.parse_args()
    f = open(configfile,'w')
    f.write(str(args.com))
    f.close()
    com = f'COM{args.com}'
    PORT = args.port
    print(f'MFCs on {com}')
    print(f'port: {PORT}')
    host = args.host
    debug = args.debug
    acceptedHosts = None
    if args.accepted_hosts:
        acceptedHosts = args.accepted_hosts.split(',')
    
    if host == 'local':
        host = 'localhost'
    elif host == 'remote':
        host = socket.gethostname()
    elif host == 'remoteip':
        host = getIP() #added this option as connection may not work with host name if host has multiple connections
    elif isValidIP(host):
        pass
    else:
        print('usage bronkorstServer [host]')
        print('host must must be "local", "remote", "remoteip" or nothing (local)')
        return
    print(host)
    return com, PORT, host, acceptedHosts, debug

def run(port = PORT):
    com, port, host, acceptedHosts = getArgs()
    
    if not acceptedHosts:
        ahstring = 'all'
    else:
        ahstring = ','.join(acceptedHosts)
        acceptedIPs = [socket.gethostbyname(h) for h in acceptedHosts]
    print(f'accepted hosts: {ahstring}')
    print('running single client server')
    mfcMain = startMfc(com)
    mfc = MFC(0,mfcMain)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind((host, port))
        s.listen()
        while True:
            conn, addr = s.accept()
            with conn:
                hostName = addr[0]
                if acceptedHosts and hostName not in acceptedHosts and hostName not in acceptedIPs:
                    print(f'{hostName} not in accepted hosts, closing connection')
                    continue
                print(f"Connected by {addr}")
                while True:
                    try:
                        data = conn.recv(1024)
                    except (ConnectionAbortedError, ConnectionResetError):
                        print('connection lost with client')
                        data = b''
                    if not data:
                        break
                    strdata = data.decode()
                    try:
                        address = int(strdata.split(';')[0])
                        mfc.address = address
                        print(strdata)
                        result = mfc.strToMethod(strdata)
                        result += '!'
                        
                    except (ValueError, KeyError):
                        byteResult = b'invalid input!'
                    resultsplit = result
                    while len(resultsplit) > 1024:
                        resultsplit = resultsplit[:1024]
                        byteResult = bytes(str(resultsplit),encoding = 'utf-8')
                        print(f'sending data to {addr}')
                        conn.sendall(byteResult)

def accept_wrapper(sock,sel):
    conn,addr =sock.accept()
    conn.setblocking(False)
    print(f'Accepted connection from {addr}')
    data = types.SimpleNamespace(addr=addr,inb = b'',outb = b'')
    events = selectors.EVENT_READ | selectors.EVENT_WRITE
    sel.register(conn,events,data=data)

def service_connection(key,mask,sel,mfc, acceptedHosts = None):
    sock = key.fileobj
    data = key.data
    #logger.debug(f'accepted connection from {data.addr}')
    connectionLostMessage = f'connection lost with client: {data.addr}'
    bytemessage = b''
    def closeConnection():
        print(f'closing connection to {data.addr}')
        sel.unregister(sock)
        sock.close()
    if acceptedHosts:
        acceptedIPs = [socket.gethostbyname(h) for h in acceptedHosts]
        hostName = data.addr[0]
        if hostName not in acceptedHosts and hostName not in acceptedIPs:
            print('host name not in accepted hosts')
            closeConnection()
            return
    if mask & selectors.EVENT_READ:
        try:
            recvData = sock.recv(1024)
            
        except (ConnectionAbortedError, ConnectionResetError):
            
            print(connectionLostMessage)
            logger.info(connectionLostMessage)
            recvData = b''
        if recvData:
            print(recvData)
            data.outb += recvData
            strmessage = data.outb.decode()
            try:
                address = int(strmessage.split(';')[0])
                mfc.address = address
                mainmessage = mfc.strToMethod(strmessage)
                #endmessageMarker = '!'
                fullmessage = f'{mainmessage}!'
                bytemessage += bytes(fullmessage,encoding='utf-8')
            except (ValueError, KeyError):
                bytemessage = b'invalid message!'
            except ConnectionResetError:
                print(f'connection lost with client: {data.addr}')
                bytemessage = b''
                closeConnection()
        else:
            logger.debug(f'no data received from {data.addr}')
            closeConnection()
    if mask & selectors.EVENT_WRITE:

        if bytemessage:
            print(f'sending data to {data.addr}')
            try:
                sent = sock.send(bytemessage)
                bytemessage = bytemessage[sent:]
                #logger.debug(f'data sent to {data.addr}')
            except ConnectionResetError:
                print(connectionLostMessage)
                logger.info(connectionLostMessage)
                bytemessage = b''
                closeConnection()

def multiServer():
    com,port, host, acceptedHosts, debug = getArgs()
    loglevel = logging.INFO
    if debug:
        loglevel = logging.DEBUG
    logging.basicConfig(filename=logfile, level = loglevel, format = '%(asctime)s %(levelname)-8s %(message)s',
                        datefmt = '%Y/%m/%d_%H:%M:%S')
    logger.info('server started')
    
    allhosts = []
    if not acceptedHosts:
        ahstring = 'all'
    else:
        ahstring = ','.join(acceptedHosts)
    print(f'accepted hosts: {ahstring}')
    mfcMain = startMfc(com)
    mfc = MFC(0,mfcMain)
    sel = selectors.DefaultSelector()
    print('running multiServer')
    s=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind((host, port))
    s.listen()
    s.settimeout(5)
    #s.setblocking(False)
    sel.register(s,selectors.EVENT_READ,data=None)

    try:
        while True:
            events = sel.select(timeout=5)
            for key, mask in events:
                if key.data is None:
                    accept_wrapper(key.fileobj,sel)
                else:
                    hostname = socket.gethostbyaddr(key.data.addr[0])
                    if not hostname in allhosts:
                        allhosts.append(hostname)
                        logger.info(f'new client {hostname}')
                    service_connection(key, mask,sel,mfc, acceptedHosts)
    except KeyboardInterrupt:
        logger.info('keyboard interupt')
        print("caught keyboard interrupt, exiting")
    except Exception as e:
        logger.exception(e)
        raise e
    finally:
        sel.close()

if __name__ == '__main__':
    run()


from bronkhorstControlbm31.bronkhorstServer import PORT, HOST, logdir
from bronkhorstControlbm31.bronkhorstClient import MFCclient
import argparse
import matplotlib.pyplot as plt
import logging, pathlib, os, time
import socket
from datetime import datetime
import numpy as np
import matplotlib
from matplotlib.widgets import CheckButtons
import pandas as pd
matplotlib.rcParams.update({'font.size':12})

homedir = pathlib.Path.home()
fulllogdir = f'{homedir}/{logdir}'
os.makedirs(fulllogdir,exist_ok=True)
logger = logging.getLogger()

def getArgs(host=HOST, port=PORT, connid = socket.gethostname(),waitTime = 0.5, plotTime = 60, log = True, logInterval = 5):
    parser = argparse.ArgumentParser()

    parser.add_argument('host',nargs='?', default=host, type= str, help = 'host name/address')
    parser.add_argument('-p','--port',default=port, type=int, help = 'port number')
    parser.add_argument('-c','--connid',default=connid, type = str, help='name for connection')
    parser.add_argument('-wt','--waittime',default=waitTime, type = float, help = 'time to wait between iterations (default 0.5 s)')
    parser.add_argument('-pt','--plotTime',default=plotTime, type = float, 
                        help = 'timePlot only. Total time to plot on x-axis (only for timePlot, default 60 minutes)')
    parser.add_argument('-l','--log', default = log, type = bool, 
                        help='timePlot only, boolean. Whether or not to log the data (default True, file saved in <homedir>/bronkhorstClientLog/<yyyymmdd>.log)')
    parser.add_argument('-li', '--logInterval', default = logInterval, type = int, help='timePlot only. Integer, time interval between each log entry (default 5 s)')
    args = parser.parse_args()

    host = args.host
    port = args.port
    connid = args.connid
    waitTime = args.waittime
    plotTime = args.plotTime
    log = args.log
    logInterval = args.logInterval

    print(host)
    print(port)
    print(connid)
    return host, port, connid, waitTime, plotTime, log, logInterval

def barPlotSingle(df, ax1, ax2, title1 = True, title2=True):
    '''
    basefontsize = 15
    width = 0.45
    mult = 1
    x = df.index.values
    
    fontsize = int(basefontsize-len(x)*0.7)
    p1 = ax1.bar(x , df['fSetpoint'].values, width, label = 'fSetpoint')
    ax1.bar_label(p1, fmt = '%.2f', padding = 3, fontsize = fontsize)
    p2 = ax1.bar(x+width*mult, df['fMeasure'].values, width, label = 'fMeasure')
    ax1.bar_label(p2, fmt = '%.2f', padding = 3, fontsize = fontsize)
    ax1.legend()
    ax1.set_xticks(x+width*mult*0.5, df['User tag'].values)
    if title1:
        ax1.set_ylabel('MFC/BPR flow')
    '''
    ax1.cla()
    ax2.cla()
    p1 = ax1.bar(df['User tag'].values, df['fMeasure'].values)
    p2 = ax2.bar(df['User tag'].values, df['fSetpoint'].values)
    ax1.bar_label(p1, fmt = '%.3f')
    ax2.bar_label(p2, fmt = '%.3f')
    if title1:
        ax1.set_ylabel('MFC/BPR measure')
    if title2:
        ax2.set_ylabel('MFC/BPR setpoint')

    

def barPlot(host=HOST, port = PORT,waittime = 1, multi = True, connid = f'{socket.gethostname()}barplot'):
    host,port,connid, waittime, _, _log, _li =getArgs(host=host,port=port,connid=connid, waitTime=waittime,plotTime=1, log = False)
    fig,ax = plt.subplots(2,1)

    while True:
        try:
            df = MFCclient(1,host,port,multi=multi, connid=connid).pollAll()
            barPlotSingle(df,ax[0], ax[1])
            plt.tight_layout()
            plt.show(block = False)
            plt.pause(waittime)
        except (KeyboardInterrupt, AttributeError) as e:
            print(e)
            plt.close(fig)
            return
        
def writeLog(file,string):
    f = open(file,'a')
    f.write(string)
    f.close()

def logMFCs(logfile, df, headerString):
    curtime = time.time()
    dt = datetime.fromtimestamp(curtime)
    dtstring = f'{dt.year:04d}/{dt.month:02d}/{dt.day:02d}_{dt.hour:02d}:{dt.minute:02d}:{dt.second:02d}'
    logString = f'{dtstring} {int(curtime)}'
    newHeaderString = f'datetime unixTime(s)'
    for i in df.index.values:
        name = df.loc[i]['User tag'].replace(' ','')
        newHeaderString += f' {name}Setpoint {name}Measure'
        meas = df.loc[i]['fMeasure']
        sp = df.loc[i]['fSetpoint']
        logString += f' {sp:.3f} {meas:.3f}'
    newHeaderString += '\n'
    logString += '\n'                 
    if newHeaderString != headerString:
        headerString = newHeaderString
        writeLog(logfile,headerString)
    writeLog(logfile,logString)


def timePlotSingle(df, ax, measure, tlist, xlim, colName = 'fMeasure', ylabel = 'MFC/BPR measure', title = True, xlabel = True, resetAxes = False):
    xlims = xlim*60

    userTags = df['User tag'].to_list()
    if tlist[-1] -tlist[0] > xlims:
        tlist.pop(0)
    for i in df.index.values:
        measure[i].append(df.loc[i][colName])
        while len(measure[i]) > len(tlist):
            measure[i].pop(0)
    tlistPlot = [t-tlist[-1] for t in tlist] 
    
    xlimzoom = ax.get_xbound()
    ylimzoom = ax.get_ybound()

    ax.cla()
    for a in measure:
        ax.plot(tlistPlot,measure[a],'o-',label = userTags[a],markersize = 3)
    if title:
        dt = datetime.fromtimestamp(tlist[-1])
        dtstring = f'{dt.year:04d}/{dt.month:02d}/{dt.day:02d} {dt.hour:02d}:{dt.minute:02d}:{dt.second:02d}'
        ax.set_title(f'measure, tscale: {xlim} minutes. Updated: {dtstring}')

    if len(tlist) > 2 and xlimzoom[0] > tlistPlot[2] and not resetAxes:
        ax.set_xbound(*xlimzoom)
        ax.set_ybound(*ylimzoom)
    ax.legend()
    if xlabel:
        ax.set_xlabel('t-current time (s)')
    ax.set_ylabel(ylabel)

def getLogFile():
    
    logdir = f'{homedir}/bronkhorstClientLog/'
    t = time.time()
    dt = datetime.fromtimestamp(t)
    dtstring = f'{dt.year:04d}{dt.month:02d}{dt.day:02d}'
    logfile = f'{logdir}/{dtstring}.log'
    if not os.path.exists(logdir):
        os.makedirs(logdir)
    return logfile

def logHeader(logfile, df):
    names = []
    headerString = f'datetime unixTime(s)'
    for i in df.index.values:
        name = df.loc[i]['User tag'].replace(' ','_')
        names.append(name)
        headerString += f' {name}Setpoint {name}Measure'
    headerString += '\n'
    writeLog(logfile,headerString)
    return headerString

def timePlot(host=HOST, port = PORT,waittime = 1, multi = True, connid = f'{socket.gethostname()}timePlot',xlim = 60, log = True, logInterval = 5):
    host,port,connid, waittime, xlim, log, logInterval = getArgs(host=host,port=port,connid=connid, waitTime=waittime,plotTime=xlim, log = log, logInterval=logInterval)
    measure = {}
    c=0
    fig,ax = plt.subplots()
    tlist = []
    if log:
        logfile = getLogFile()
    tlog = 0

    while True:
        try:
            tlist.append(time.time())
            df = MFCclient(1,host,port,multi=multi, connid=connid).pollAll()
            df = df.astype({'fMeasure':float})
            if c == 0:
                for i in df.index.values:
                    measure[i] = []
                c = 1
                if log:
                    headerString = logHeader(logfile, df)
            timePlotSingle(df,ax,measure, tlist, xlim)

            if log and time.time() - tlog > logInterval:
                logMFCs(logfile,df, headerString)
                tlog = time.time()
                
            plt.tight_layout()
            plt.show(block = False)
            plt.pause(waittime)
        except (KeyboardInterrupt,AttributeError):
            plt.close(fig)
            return
        
def plotValvesBar(df, ax):
    p1 = ax.bar(df['User tag'].values, df['Valve output'].values)
    ax.bar_label(p1, fmt = '%.2f')
    ax.set_ylabel('MFC/BPR Measure')


class Plotter():
    def __init__(self,host=HOST, port = PORT,waittime = 1, connid = f'{socket.gethostname()}allPlot',xlim = 60, log = True, logInterval = 5):
        self.host = host
        self.port = port
        self.waittime = waittime
        self.connid = connid
        self.xlim = xlim
        self.log = log
        self.logInterval = logInterval
        self.fig, self.ax = plt.subplots(2,2)
        self.measureFlow = {}
        self.measureValve = {}
        self.tlist = []
        
        self.logfile = getLogFile()
        self.tlog = 0
        df = MFCclient(1,self.host,self.port, connid=self.connid).pollAll()
        if self.log:
            self.headerString = logHeader(self.logfile,df)
        axes = plt.axes([0.001, 0.0001, 0.08, 0.05])
        axes.axis('off')
        self.radiobutton = CheckButtons(axes, ['reset axes'],[False])
        for i in df.index.values:
            self.measureFlow[i] = []
            self.measureValve[i] = []
        self.fig.show()
    def plotAll(self):
        eventlogfile = f'{homedir}/{logdir}/mfcPlotAll.log'
        logging.basicConfig(filename=eventlogfile, level = logging.INFO, format = '%(asctime)s %(levelname)-8s %(message)s',
                            datefmt = '%Y/%m/%d_%H:%M:%S')
        logger.info('mfcPlotAll started')
        plt.ion()
        
        while True:
            try:
                df = MFCclient(1,self.host,self.port, connid=self.connid).pollAll()
                self.plotAllSingle(df)
                plt.pause(self.waittime)
                
            except KeyboardInterrupt:
                logger.info('keyboard interrupt')
                plt.close(self.fig)
                return
            except AttributeError as e:
                logger.error(f'{e}, possible keyboard interrupt during connection')
                return
            except OSError as e:
                message = f'{e}.\nbronkhorstServer is probably not open, or host or port settings are incorrect'
                print(message)
                logger.error(message)
                return
            except ConnectionResetError as e:
                message = f'{e}.\nbronkhorstServer likely closed while running'
                logger.error(message)
                print(message)
                return
            except np.core._exceptions._UFuncNoLoopError as e:
                logger.warning(e)
                continue
            except Exception as e:
                logger.exception(e)
                raise e
    def plotAllSingle(self,df):
        plt.ion()
        self.tlist.append(time.time())
        self.resetAxes = self.radiobutton.get_status()[0]

        barPlotSingle(df,self.ax[0,1], self.ax[1,1], title1=True)

        timePlotSingle(df,self.ax[0,0],self.measureFlow,self.tlist,self.xlim, xlabel=True, resetAxes=self.resetAxes)

        if self.log and time.time() - self.tlog > self.logInterval:
            logMFCs(self.logfile, df, self.headerString)
            self.tlog = time.time()

        timePlotSingle(df,self.ax[1,0], self.measureValve, self.tlist, self.xlim, colName='Valve output', ylabel='MFC/BPR valve output',
                        title=False, resetAxes=self.resetAxes)

        plt.subplots_adjust(top = 0.95, bottom = 0.07, right = 0.99, left = 0.1, 
            hspace = 0.2, wspace = 0.2)


            
def plotAll(host = HOST, port = PORT, connid = f'{socket.gethostname()}allPlot', xlim = 60, log = True, logInterval = 5, waittime = 0.5):
    host,port,connid,waittime, xlim, log, logInterval = getArgs(host=host,port=port,connid=connid, waitTime=waittime,plotTime=xlim, 
                                                                log = log, logInterval=logInterval)
    plotter = Plotter(host=host, port = port,waittime = waittime, connid = connid,xlim = xlim, log = log, logInterval = logInterval)
    plotter.plotAll()
